#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <GL/glpng.h>

#include "texture.h"
#include "laby.h"

static int imgflen = 32;
static int tex_num = 32;
static int sw_num;
static int unused;

static struct wall_img{
  int i;
  char rot;
  char flip;
  GLuint tex;
} *imgfield;

static int getsw(char side, int a, int b)
{
  int A,B,r;

  switch(side/2){
  case 0: A = height; B = depth;  r = 0; break;
  case 1: A = width;  B = depth;  r = 2*height*depth; break;
  case 2: A = width;  B = height; r = 2*(height+width)*depth; break;
  }
  if(side%2) r += A*B;
  return r+a+A*b;
}

void putimages()
{
  int i, j, j_max, k, wall_num, pom, sc0, sc1, sides;
  int *walls;

  if(!texture) return;

  //imgflen = width*height*depth;

  wall_num = width*height*depth*4;

  if(imgflen > wall_num){
    warning("too many images");
    imgflen = wall_num;
  }

  imgfield = (struct wall_img *)malloc(sizeof(struct wall_img)*imgflen);
  if(!imgfield){
    warning("imgfield allocation failed");
    return;
  }

  sc0 = getsw(SCside[0], SCA[0], SCB[0]);
  sc1 = getsw(SCside[1], SCA[1], SCB[1]);
  if(sc1 < sc0){
    pom = sc0;
    sc0 = sc1;
    sc1 = pom;
  }

  sides = 2*(width*height+width*depth+depth*height);

  for(i=0; i<imgflen; i++){
    imgfield[i].i = rand()%(wall_num-i);
    if(unused){
      j = rand()%unused;
      imgfield[i].tex = texture[j];
      unused--;
      texture[j] = texture[unused];
      texture[unused] = imgfield[i].tex;
    }
    else imgfield[i].tex = texture[rand()%tex_num];
    if(imgfield[i].i >= sc0) imgfield[i].i++;
    if(imgfield[i].i >= sc1) imgfield[i].i++;
    for(j=0; j<i && imgfield[j].i <= imgfield[i].i; j++){
      imgfield[i].i++;
      if(imgfield[i].i == sc0) imgfield[i].i++;
      if(imgfield[i].i == sc1) imgfield[i].i++;
    }
    pom = imgfield[i].i;
    for(k=i; k>j; k--) imgfield[k].i = imgfield[k-1].i;
    imgfield[j].i = pom;
  }

  sw_num = 0;

  j = 0;
  walls = walls_x;
  j_max = (width-1)*height*depth;
  k = sides/2;

  for(i=0; i<imgflen; i++){
    imgfield[i].rot = rand()%4;
    if(imgfield[i].i < sides){
      imgfield[i].flip = 0;
      if(imgfield[i].i < height*depth) imgfield[i].flip = 1;
      else if(imgfield[i].i >= 2*height*depth && imgfield[i].i < (2*height+width)*depth) imgfield[i].flip = 1;
      else if(imgfield[i].i >= 2*(height+width)*depth + height*width) imgfield[i].flip = 1;
      sw_num++;
    }
    else{
      while(k != imgfield[i].i/2 || !walls[j]){
	if(walls[j]) k++;

	j++;
	if(j == j_max){
	  if(walls == walls_z) break;
	  j=0;
	  if(walls == walls_x){
	    walls = walls_y;
	    j_max = (height-1)*width*depth;
	  }
	  else{
	    walls = walls_z;
	    j_max = (depth-1)*width*height;
	  }
	}
      }
      if(j == j_max){
	warning("dosly mi zdi: %d, %d, %d", sides/2, k, imgfield[i].i/2);
	return;
      }
      imgfield[i].flip = imgfield[i].i%2;
      if(walls[j] > 1) imgfield[i-1].flip = 2;
      else walls[j] = 2+i;
    }
  }
}

static void loadtex(char *filename, GLuint *save)
{
  pngInfo info;

  glGenTextures(1, save);
  glBindTexture(GL_TEXTURE_2D, *save);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

  if (pngLoad(filename, PNG_NOMIPMAP, PNG_SOLID, &info)) {
    /*
    printf("Loaded %s with resounding success\n", filename);
    printf("Size=%i,%i Depth=%i Alpha=%i\n", info.Width, info.Height, info.Depth, info.Alpha);
    */
  }
  else error("Can't load %s", filename);
}

void inittex()
{
  char filename[30];
  int i;
  GLubyte white[4] = {255, 255, 255, 255};

  glEnable( GL_TEXTURE_2D );

  glGenTextures(1, &emptytex);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, &white);

  loadtex("vchod.png", &SCtex);
  texture = (GLuint *)malloc(sizeof(GLuint)*tex_num);
  if(!texture){
    warning("textures mallocation failed");
    return;
  }
  for(i=0; i<tex_num; i++){
    sprintf(filename, "texture%d.png", i);
    loadtex(filename, &texture[i]);
  }
  unused = tex_num;
}

void drawsquare(int x, int y, int z, char dir, char flip)
{
  int w;

  w = iswall(x,y,z,dir);
  if(!w) return;
  if(w == 1) glRectf(-0.5, -0.5, 0.5, 0.5);
  else{
    w -= 2;
    if(flip+imgfield[w].flip == 3) w++;
    if(flip == imgfield[w].flip%2)
      texturect(imgfield[w].tex, imgfield[w].rot, flip);
    else glRectf(-0.5, -0.5, 0.5, 0.5);
  }
}

void texturect(int tex, char rot, char flip)
{
  int a = flip;
  int b = 1-flip;
  int pom;

  if(rot/2){
    a = 3-a;
    b = 3-b;
  }
  if(rot%2){
    pom = a; a = b; b = 3-pom;
  }

  glBindTexture(GL_TEXTURE_2D, tex);
  glBegin( GL_QUADS );
  glTexCoord2i(a%2, a/2); glVertex2d(-0.5+wall_border, -0.5+wall_border);
  glTexCoord2i(b%2, b/2); glVertex2d(+0.5-wall_border, -0.5+wall_border);
  glTexCoord2i((3-a)%2, (3-a)/2); glVertex2d(+0.5-wall_border, +0.5-wall_border);
  glTexCoord2i((3-b)%2, (3-b)/2); glVertex2d(-0.5+wall_border, +0.5-wall_border);
  glEnd();

  glBindTexture(GL_TEXTURE_2D, emptytex);
}

int iswall(int x, int y, int z, char dir)
{
  int *wall;
  int i, j;

  wall = wall_coor(x,y,z, dir);
  if(!wall){
    i = 0;
    if(x>=0 && y>=0 && z>=0 && x<width && y<height && z<depth) i = 1;
    movecoor(&x, &y, &z, dir);
    if(x>=0 && y>=0 && z>=0 && x<width && y<height && z<depth){
      dir = backdir(dir);
      i = 1;
    }

    if(i && imgfield){
      if(dir/2 == 0) i = getsw(dir, y, z);
      else if(dir/2 == 1) i = getsw(dir, x, z);
      else i = getsw(dir, x, y);

      for(j=0; j < sw_num; j++) if(imgfield[j].i == i) return j+2;
      return 1;
    }
    return 0;
  }
  return *wall;
}
